import argparse
import gzip
import os
from typing import Callable, Union

from danmakuC.__version__ import __version__
# sites
from danmakuC import bilibili, niconico


def is_gzip_file(filename: Union[str, os.PathLike]) -> bool:
    gzip_magic = b'\x1f\x8b'
    with open(filename, mode='rb') as fp:
        return fp.read(2) == gzip_magic


def get_convert_func(proto_file: Union[str, os.PathLike], ) -> Callable:
    if is_gzip_file(proto_file):
        open_func = gzip.open
    else:
        open_func = open
    with open_func(proto_file, mode='rb') as fp:
        firstbyte = fp.read(1)
    if firstbyte == b'\x0a':
        convert_func = bilibili.proto2ass
    else:
        convert_func = niconico.proto2ass
    return convert_func


def main():
    parser = argparse.ArgumentParser(description=f"danmakuC cli version {__version__}", prog="danmakuC")
    parser.add_argument("file", help="Comment file to be processed")
    parser.add_argument("-o", "--output", help="Output file")

    parser.add_argument("-s", "--size", type=str, help="Stage size in pixels [default: 1920x1080]", default="1920x1080")
    parser.add_argument(
        "-rb", "--reserve-blank", help="Reserve blank on the bottom of the stage [default: 0]", type=int, default=0)
    parser.add_argument(
        "-fn",
        "--font",
        help="Specify font face [default: sans-serif]",
        default="sans-serif",
    )
    parser.add_argument(
        "-fs",
        "--fontsize",
        help="Default font size [default: 25.0]",
        type=float,
        default=25.0,
    )
    parser.add_argument("-a", "--alpha", help="Alpha [default: 1.0]", type=float, default=1.0)
    parser.add_argument(
        "-dm",
        "--duration-marquee",
        help="Duration of scrolling comment display [default: 5.0]",
        type=float,
        default=5.0,
    )
    parser.add_argument(
        "-ds",
        "--duration-still",
        help="Duration of still comment display [default: 5.0]",
        type=float,
        default=5.0,
    )
    parser.add_argument("-fl", "--filter", help="Regular expression to filter comments", default="")
    # parser.add_argument(
    #     "-flf", "--filter-file", help="Regular expressions from file (one line one regex) to filter comments"
    # )
    parser.add_argument("-r", "--reduce", action="store_true", help="Reduce the amount of comments if stage is full")
    parser.add_argument("-v", "--version", action="version", version=f"version {__version__}")
    # parse args
    args = parser.parse_args()
    width, height = args.size.split("x")
    width = int(width)
    height = int(height)

    with open(args.file, 'rb') as f:
        res = get_convert_func(args.file)(
            f,
            width,
            height,
            args.reserve_blank,
            args.font,
            args.fontsize,
            args.alpha,
            args.duration_marquee,
            args.duration_still,
            args.filter,
            args.reduce,
            out_filename=args.output,
        )
    if res is not None:
        print(res)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nndcomment.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10nndcomment.proto\"\xd6\x01\n\nNNDComment\x12\x0e\n\x06thread\x18\x01 \x01(\x03\x12\n\n\x02no\x18\x02 \x01(\x05\x12\x0c\n\x04vpos\x18\x03 \x01(\x05\x12\x0c\n\x04\x64\x61te\x18\x04 \x01(\x03\x12\x11\n\tdate_usec\x18\x05 \x01(\x05\x12\x11\n\tanonymity\x18\x06 \x01(\x08\x12\x0f\n\x07user_id\x18\x07 \x01(\t\x12\x0c\n\x04mail\x18\x08 \x01(\t\x12\x0c\n\x04leaf\x18\t \x01(\x05\x12\x0f\n\x07premium\x18\n \x01(\x08\x12\r\n\x05score\x18\x0b \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x0c \x01(\t\x12\x0c\n\x04\x66ork\x18\x0f \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nndcomment_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _NNDCOMMENT._serialized_start=21
  _NNDCOMMENT._serialized_end=235
# @@protoc_insertion_point(module_scope)



import json
from google.protobuf.json_format import MessageToJson
from .ass import Ass
from .protobuf.bilibili import BiliCommentProto, BiliViewProto
from typing import Union, Optional
import io

__all__ = ['proto2ass', 'parse_view']


def proto2ass(
        proto_file: Union[bytes, io.IOBase],
        width: int,
        height: int,
        reserve_blank: int = 0,
        font_face: str = "sans-serif",
        font_size: float = 25.0,
        alpha: float = 1.0,
        duration_marquee: float = 5.0,
        duration_still: float = 5.0,
        comment_filter: str = "",
        reduced: bool = False,
        out_filename: str = "",
) -> Optional[str]:
    ass = Ass(width, height, reserve_blank, font_face, font_size, alpha, duration_marquee,
              duration_still, comment_filter, reduced)

    if isinstance(proto_file, io.IOBase):
        proto_file = proto_file.read()
    target = BiliCommentProto()
    target.ParseFromString(proto_file)
    for elem in target.elems:
        if elem.mode == 8:
            continue  # ignore scripted comment
        try:
            ass.add_comment(
                elem.progress / 1000,  # 视频内出现的时间
                elem.ctime,  # 弹幕的发送时间（时间戳）
                elem.content,
                elem.fontsize,
                {1: 0, 4: 2, 5: 1, 6: 3, 7: 4}[elem.mode],
                elem.color,
            )
        except TypeError:  # incase integer overflow https://github.com/HFrost0/bilix/issues/102
            continue
    if out_filename:
        return ass.write_to_file(out_filename)
    else:
        return ass.to_string()


def parse_view(content: bytes) -> dict:
    dm_view = BiliViewProto()
    dm_view.ParseFromString(content)
    dm_view = json.loads(MessageToJson(dm_view))
    return dm_view



import time
from biliass import Danmaku2ASS
from danmakuC.bilibili import proto2ass

file = 'test_dm_large.bin'


def test_cmp():
    """"
    test_dm.bin and test_dm_large.bin is obtained by bilix from bilibili.com
    """
    text = test_new()
    text2 = test_old()
    for l1, l2 in zip(text.split('\n'), text2.split('\n')):
        # todo different time convert due to ceil ?
        if l1 != l2:
            print(l1)
            print(l2)


def test_new():
    a = time.time()
    with open(file, "rb") as f:
        text = proto2ass(f.read(), 1920, 1080)
    print(time.time() - a)
    return text


def test_old():
    a = time.time()
    with open(file, "rb") as f:
        text = Danmaku2ASS(f.read(), 1920, 1080, input_format="protobuf")
    print(time.time() - a)
    return text



# Defined in csrc/ass.cpp

class Ass:
    def __init__(self,
                 width: int,
                 height: int,
                 reserve_blank: int = 0,
                 font_face: str = "sans-serif",
                 font_size: float = 25.0,
                 alpha: float = 1.0,
                 duration_marquee: float = 5.0,
                 duration_still: float = 5.0,
                 filter: str = "",
                 reduced: bool = False): ...

    def add_comment(self, progress: float, ctime: int, content: str, font_size: float, mode: int, color: int) -> bool:
        """add a comment to Ass object, return True if add success"""
        ...

    def to_string(self) -> str:
        """return the ass text"""
        ...

    def write_to_file(self, str) -> None:
        """direct write to the file to avoid memory cost"""



from pybind11.setup_helpers import build_ext, Pybind11Extension
from setuptools import setup, find_packages

ext_modules = [
    Pybind11Extension(
        "danmakuC.ass",
        [
            'danmakuC/csrc/ass.cpp',
            'third_party/fmt/src/format.cc',
        ],
        include_dirs=[
            'third_party/boost_1_81_0',
            'third_party/fmt/include',
        ],
    ),
]


def get_version() -> str:
    with open('danmakuC/__version__.py', 'r', encoding='utf8') as f:
        for line in f.readlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


def get_long_description():
    with open('README.md', 'r', encoding='utf8') as f:
        return f.read()


setup(
    name="danmakuC",
    version=get_version(),
    description="Faster conversion for larger Danmaku to Ass format",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: C++',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
    author="HFrost0, m13253, Nyakku Shigure",
    license="GPLv3",
    packages=find_packages(exclude=['tests', 'danmaku_old']),
    package_data={'danmakuC': ["*.pyi"]},
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    # https://cibuildwheel.readthedocs.io/en/stable/faq/#windows-importerror-dll-load-failed-the-specific-module-could-not-be-found
    python_requires=">=3.8.3",
    install_requires=["protobuf>=4.21.0"],
    entry_points={
        "console_scripts": "danmakuC=danmakuC.__main__:main",
    },
)



import io
import re
from .ass import Ass
from .protobuf.niconico import NNDCommentProto
from typing import Union, Optional

__all__ = ['proto2ass']


def proto2ass(
        proto_file: Union[bytes, io.IOBase],
        width: int,
        height: int,
        reserve_blank: int = 0,
        font_face: str = "sans-serif",
        font_size: float = 25.0,
        alpha: float = 1.0,
        duration_marquee: float = 5.0,
        duration_still: float = 5.0,
        comment_filter: str = "",
        reduced: bool = False,
        out_filename: str = "",
) -> Optional[str]:
    ass = Ass(width, height, reserve_blank, font_face, font_size, alpha, duration_marquee,
              duration_still, comment_filter, reduced)
    if isinstance(proto_file, bytes):
        proto_file = io.BytesIO(proto_file)
    comment = NNDCommentProto()
    while True:
        size = int.from_bytes(proto_file.read(4), byteorder='big', signed=False)
        if size == 0:
            break
        comment_serialized = proto_file.read(size)
        comment.ParseFromString(comment_serialized)
        pos, color, size = process_mailstyle(comment.mail, font_size)
        ass.add_comment(
            comment.vpos / 100,
            comment.date,
            comment.content,
            size,
            pos,
            color,
        )
    if out_filename:
        return ass.write_to_file(out_filename)
    else:
        return ass.to_string()


def process_mailstyle(mail, fontsize):
    pos, color, size, patissier = 0, 0xffffff, fontsize, False
    if not mail:
        return pos, color, size  # , patissier
    for mailstyle in mail.split():
        if mailstyle == 'ue':  # top middle
            pos = 1
        elif mailstyle == 'shita':  # bottom middle
            pos = 2
        elif mailstyle == 'naka':  # flying left-to-right
            pos = 0
        elif mailstyle == 'big':
            size = fontsize * 1.44
        elif mailstyle == 'small':
            size = fontsize * 0.64
        elif mailstyle in NICONICO_COLOR_MAPPINGS:
            color = NICONICO_COLOR_MAPPINGS[mailstyle]
        elif len(mailstyle) == 7 and re.match(HEX_COLOR_REGEX, mailstyle):
            color = int(re.match(HEX_COLOR_REGEX, mailstyle).group(1), base=16)
        elif mailstyle == 'patissier':  # for comment art/fixed speed?
            patissier = True

    return pos, color, size  # , patissier


# https://w.atwiki.jp/nicoapi/pages/20.html
NICONICO_COLOR_MAPPINGS = {
    # Regular users
    'red': 0xff0000,
    'pink': 0xff8080,
    'orange': 0xffcc00,
    'yellow': 0xffff00,
    'green': 0x00ff00,
    'cyan': 0x00ffff,
    'blue': 0x0000ff,
    'purple': 0xc000ff,
    'black': 0x000000,
    # Premium users
    'niconicowhite': 0xcccc99,
    'white2': 0xcccc99,
    'truered': 0xcc0033,
    'red2': 0xcc0033,
    'passionorange': 0xff6600,
    'orange2': 0xff6600,
    'madyellow': 0x999900,
    'yellow2': 0x999900,
    'elementalgreen': 0x00cc66,
    'green2': 0x00cc66,
    'marineblue': 0x33ffcc,
    'blue2': 0x33ffcc,
    'nobleviolet': 0x6633cc,
    'purple2': 0x6633cc,
}

HEX_COLOR_REGEX = re.compile('#([a-fA-F0-9]{6})')
